from django.apps import AppConfig
from django.contrib.auth.hashers import check_password, make_password, is_password_usable
from django.db import models
from sculpt.common import Enumeration
from sculpt.model_tools.hash_generator import ModelHashGenerator
import datetime

# Useful things to include in Model definitions

# a really simple convert-to-unicode method which
# is better than Django's default; good for
# prototyping, but a real class needs more
#
class SimpleUnicodeMixin(object):

    def __unicode__(self):
        if hasattr(self, 'title'):
            return u'[%(cls)s:%(id)s] %(title)s' % {
                    'cls': self.__class__.__name__,
                    'id': str(self.id),     # in case id is None, don't barf
                    'title': self.title,
                }
        else:
            return u'[%(cls)s:%(id)s]' % {
                    'cls': self.__class__.__name__,
                    'id': str(self.id),     # in case id is None, don't barf
                }

# AUTO-GENERATED HASH ID MODEL
#
# Often we want to expose a record ID to end users
# but doing so using Django's built-in id field has
# some disadvantages:
#
#   1. It exposes how many of a thing were present
#      in the system prior to that object's creation
#      (a business intelligence leak)
#
#   2. It is enumerable: it is trivial to write a
#      script that checks every possible ID, looking
#      for things that were inadequately protected
#      or are otherwise interesting (a privacy leak)
#
# The solution is to generate a unique hash, spreading
# the sequential id values "randomly" over a 256-bit
# hash space. This can then be returned to the user,
# as even with four billion entries in the system
# (2^32) there is only a 1 in 6.277e57 chance of any
# randomly-chosen hash matching a record in the system.
# (Roughly.)
#
# Some problems arise:
#
#   1. The hash must be unique, but the ID value is
#      typically not available when the hash needs to
#      be generated because it's a new record. This
#      means the hash MUST be generated from other
#      data, typically from the object, as well as the
#      datetime.
#
#   2. Generated hashes must be checked against the
#      database for uniqueness prior to being written;
#      this means there is a race condition where two
#      concurrent inserts of identical data may clash
#      and cause one to fail, even with the additional
#      checks.
#
#   3. Since hash generation data comes from outside
#      the field itself, this is a MODEL mix-in and not
#      just a simple field.
#
# You will need to define some additional values in the
# model in order for this to work:
#
#   AUTOHASH_SECRET - a string with unique data so that
#       hashes from different models aren't the same
#       (required)
#
#   AUTOHASH_FIELDS - a list of field names which will
#       be converted to strings, concatenated, and hashed
#       as one (required)
#
#   AUTOHASH_NO_DATETIME - if set to True, the current
#       datetime will not automatically be included in
#       the hash generation (default: False)
#
#   AUTOHASH_ALLOW_EMPTY - if set to False, empty or null
#       hash values will be populated before saving;
#       otherwise, they will only be generated when
#       generate_hash is specifically called (default: False)
#
class AutoHashMixin(object):
    
    # a convenient wrapper around the base ModelHashGenerator,
    # which automatically fetches all the input fields and
    # hands them off to be incorporated into the hash
    def generate_hash(self):
    
        # collect all the arguments together
        args = [ getattr(self, field) for field in self.AUTOHASH_FIELDS ]

        # include current timestamp unless we're directed not to
        if not getattr(self, 'AUTOHASH_NO_DATETIME', False):
            args.append(datetime.datetime.utcnow())

        # AUTOHASH_SECRET is required
        if self.AUTOHASH_SECRET is None:
            raise Exception('AUTOHASH_SECRET must be defined in your derived class. Refusing to operate without a defined secret.')

        # generate the hash and record it            
        self.hash = ModelHashGenerator.generate_hash(self.__class__, self.AUTOHASH_SECRET, *args)
        
    # override the model save method
    def save(self, *args, **kwargs):
    
        # if we're not allowing empty hashes, and this object
        # has an empty hash, fill it in right now
        if not getattr(self, 'AUTOHASH_ALLOW_EMPTY', False) and (self.hash == None or self.hash == ''):
            self.generate_hash()
            
        # pass through to the regular save method
        return super(AutoHashMixin, self).save(**kwargs)


# a URL pattern fragment to match hashes
AUTOHASH_URL_PATTERN = r'[-_0-9A-Za-z]{43}'
 
# LoginMixin
#
# Logging in doesn't have anything to do with authentication
# (determining who the user is) or authorization (determing
# whether a user is allowed to log in). Those steps should
# already be complete before logging in occurs. Logging in
# is all about recording an "active" user for a particular
# browser session, and logging out is all about breaking that
# association and scrubbing the session of any personalized
# data.
#
# Add this to a model to give it helper functions to work
# with request and session objects to manage the login
# process.
#
# OPTIONAL OVERRIDES:
#   LOGIN_ID_KEY - a string that is the key used to store
#       the ID of the user record into session
#   LOGIN_REQUEST_KEY - a string that is the Key used to
#       store the User instance into the request object
#
# NOTE: we define these two to be different from what
# Django's internal user management would be so that we can
# be logged in with both an app user and a Django user in
# the same session.
#
class LoginMixin(object):
    
    LOGIN_ID_KEY = 'app_user_id'
    LOGIN_REQUEST_KEY = 'app_user'

    # Sets the appropriate values in the session to stay logged in.
    def login(self, request):
        # Django Bug Fix
        # This is to force session_cache to load if it's not
        # already on the session object.
        request.session._get_session()

        # Enforce that you start at a clean slate
        request.session.cycle_key()
        request.session[self.LOGIN_ID_KEY] = self.pk

        # Cache the logged-in user object in the request (for
        # the remainder of this request)
        setattr(request, self.LOGIN_REQUEST_KEY, self)

    # In our login we are setting a key onto the request object
    # itself, and we need to clear that out, then call the super
    # which should get rid of the session
    @classmethod
    def logout(cls, request):
        setattr(request, cls.LOGIN_REQUEST_KEY, None)
        request.session.flush()

    # determine whether a session is logged in without making
    # an additional database request; we just test the ID of
    # the user record in the session
    @classmethod
    def is_logged_in(cls, request):
        value = False
        if cls.LOGIN_ID_KEY in request.session and cls.get_login_user_id(request) != None:
            value = True
        return value

    # fetch the currently-active user for a request
    #
    # This is checked first in the request object itself (in case
    # the user has already been fetched) and then, if an ID is
    # present in session but no user has been fetched, fetch it
    # from a database.
    #
    # NOTE: we use the default REQUEST and ID keys (attribute
    # names) to make life simple, but if you have an application
    # with multiple user classes (e.g. app user, back office user)
    # you will want to have different keys for these so that they
    # do not have any chance of overlapping. This allows you to
    # do easy checks (if Appuser.get_login_user(request) == None)
    # instead of constantly having to check the type of a returned
    # user, and allows you to add new non-overlapping classes of
    # users without revisiting existing code.
    #
    @classmethod
    def get_login_user(cls, request):
        # if we've already written the app_user to the request,
        # go ahead and return it without requiring an additional
        # database fetch
        if hasattr(request, cls.LOGIN_REQUEST_KEY):
            return getattr(request, cls.LOGIN_REQUEST_KEY)

        # Default the app_user to None if you are not logged in. The
        # following test could fail (because the session got flushed
        # but the request object's user was not removed) so we are
        # paranoid. Our own logout code DOES remove the object but
        # we'd rather be sure.
        setattr(request, cls.LOGIN_REQUEST_KEY, None)
        if cls.is_logged_in(request = request):
            # get the user if it exists, otherwise it's None;
            # cache the result in the request object
            setattr(request, cls.LOGIN_REQUEST_KEY, cls.get_login_user_queryset(request).first())
            return getattr(request, cls.LOGIN_REQUEST_KEY)

    # generate the queryset used to select users
    #
    # This allows you to customize the queryset without rewriting
    # all the login logic (e.g. if you wanted to always fetch some
    # related records, or apply other rules).
    #
    # It also gives you a base queryset for further expansion.
    #
    @classmethod
    def get_login_user_queryset(cls, request):
        return cls.objects.filter(pk = cls.get_login_user_id(request))

    # fetch just the ID of the logged-in user, if any
    # (automatically uses the correct key)
    @classmethod
    def get_login_user_id(cls, request):
        return request.session.get(cls.LOGIN_ID_KEY)

# PasswordMixin
#
# Includes functions that are useful in managing passwords
# that are stored locally as one-way hashes. This is
# intended for use in models derived from AbstractAppUserCredential.
# AbstractSimpleAppUser does not need this class because it
# derives from AbstractBaseUser, from which this is liberally
# cribbed.
#
# OPTIONAL OVERRIDES:
#   PASSWORD_FIELD - field name that contains the password hash;
#       defaults to "password" but should be overridden to "data2"
#       for AbstractAppUserCredential-derived classes.
#
# NOTE: From time to time Django updates the hashing functions used
# for passwords to make them stronger whenever weaknesses are found.
# However, since passwords are hashed it's not possible to upgrade
# the strength of all the password hashes at once, because the
# plaintext version of the password is not available. Django thus
# uses an opportunistic approach: whenever a password is tested,
# and a confirmed match is found, it will be updated in the database
# if it needs to be upgraded to a stronger algorithm. For this
# reason the check_password method MAY update the record behind the
# scenes. The set_password method, however, NEVER does this; you
# must explicitly save() the record after setting the password.
#
class PasswordMixin(object):

    PASSWORD_FIELD = 'password'

    def set_password(self, raw_password):
        setattr(self, self.PASSWORD_FIELD, make_password(raw_password))

    def check_password(self, raw_password):
        """
        Returns a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """
        def setter(raw_password):
            self.set_password(raw_password)
            self.save(update_fields=[self.PASSWORD_FIELD])
        return check_password(raw_password, getattr(self, self.PASSWORD_FIELD), setter)

    def set_unusable_password(self):
        # Sets a value that will never be a valid hash
        setattr(self, self.PASSWORD_FIELD, make_password(None))

    def has_usable_password(self):
        return is_password_usable(getattr(self, self.PASSWORD_FIELD))

    # convenience wrapper, so passwords can be made during
    # object creation
    @classmethod
    def make_password(cls, raw_password):
        return make_password(raw_password)

# OverridableChoices
#
# Often in abstract base classes we need to provide an Enumeration
# of choices for a particular field, but we expect or require the
# concrete implementation to override this with an app-specific
# list. Django gets a bit confused by this because of how it builds
# fields incrementally from base classes on up. This is unhelpful.
#
# To fix this, we could simply reset the choice list every time we
# instantiate an object. This works, but Django's admin "add" pages
# generate their form based on the class object, not an instance
# object, so they won't see the updated choice list.
#
# It seems easy to just modify the choice list in the class object
# after it's been defined. The problem is that the mechanism for
# fetching the field (Foo._meta.get_field_by_name) isn't usable
# until the app registry is finished. Reproducing the logic here
# makes no sense.
#
# Instead, we take note of the request to update the model fields,
# listen for the class_prepared signal, and then update it. So,
# after a class is defined, call the method to register the update
# request, and it will happen automagically (ugh) once Django gets
# around to finishing the model loading.
#
#   class Foo(AbstractBaseClass):
#       ENUMERATION_NAME = Enumeration( ... )
#
#   Foo._register_field_choices('my_field', Foo.ENUMERATION_NAME)
#
# We call _register_field_choices for each affected field. The
# function itself is included in this mix-in.
#
# NOTE: you only have to do this if you actually override the
# enumeration from the base class's version. IF it's unchanged,
# you can skip this step.
#
# NOTE: YOU MUST ADD sculpt.model_tools TO YOUR INSTALLED_APPS
# SETTING or Django never finds the bit of code that acts on
# the registered changes after the app registry (and model
# loading) is complete.
#
class OverridableChoicesMixin(object):

    # This is a magical function.  In Django's internals, 
    # you can pull out a field and set various amount of data on it.
    # One use case would be to overrride the set choices from this 
    # abstract class with the values in the concrete class.
    # 
    # self will be your instance, a subclass of this class
    # self._meta is a piece of django's underbelly that 
    #     keeps the fields for a model
    # self.meta.get_field_by_name is a function that pulls out a tuple
    #     that looks like (<field>, None, Yes, No), which is why we ask 
    #     for [0] to get the field
    # self._meta.get_field_by_name(field_name)[0]._choices is the field 
    #     we want to override.  It has the values that the field is going
    #     to validate against when asked to clean.
    #
    # NOTE: we accept the actual choices OR an Enumeration.
    #
    @classmethod
    def _set_field_choices(cls, field_name, choices):
        if isinstance(choices, Enumeration):
            choices = choices.choices
        cls._meta.get_field_by_name(field_name)[0]._choices = choices

    @classmethod
    def _register_field_choices(cls, field_name, choices):
        field_choices[cls] = (field_name, choices)
    
# the signal handler that updates choices after classes are prepared    
field_choices = {}

class OverridableChoicesConfig(AppConfig):

    name = "sculpt.model_tools"
    verbose_name = "Code Sculpture Model Tools"

    def ready(self):
        for cls, args in field_choices.iteritems():
            cls._set_field_choices(*args)

# SimpleTreeMixin
#
# There are many ways to implement tree structures in SQL and various
# libraries for doing so. The simplest method is for each node to
# reference its parent and to include a display order value that
# applies amongst its siblings. This has the virtues of requiring only
# changes to immediate siblings when nodes are inserted or reordered
# and that moving whole sub-trees requires only a single field to be
# changed.
#
# The downside to this method is that fetching all the descendants of
# any particular node requires repeated queries to the database (one
# per generation) and fetching all the ancestors does also. There are
# schemes that optimize for that kind of query (such as the MPTT
# scheme) at the expense of increased cost of updates.
#
# We observe that the cost of updates, especially with large tree
# structures which may have multiple actively-modified areas at any
# given time, is nothing to sneer at. (Inserting a node and requiring
# 10,000 other nodes to be updated seems... unwise.) We also observe
# that with large trees it is unlikely that the whole tree would be
# presented to the user anyway, so being able to selectively limit
# the number of generations of children fetched is actually a good
# thing.
#
# To use this, you will need to define two fields within your class:
#
#   parent = models.ForeignKey('self', related_name = 'children', blank = True, null = True)
#   display_order = models.IntegerField(default = 0)
#
class SimpleTreeMixin(object):

    # get all ancestors, starting with the closest; if
    # oldest_first is True, returns the farthest ancestor
    # first instead (and it will return an iterator
    # instead of a bare list, use list() if you need to)
    #
    # NOTE: because of the way children fetching works, if
    # you locate the parents and then fetch the children
    # you will not get the same Python object that you
    # started with; you will get a copy of the same data
    # from the database in a new Python object.
    #
    def get_parents(self, oldest_first = False, stop_at_id = None):
        ancestors = []
        node = self
        while node.parent is not None and (stop_at_id is None or stop_at_id != node.pk):
            node = node.parent
            ancestors.append(node)

        if oldest_first:
            return reversed(ancestors)
        else:
            return ancestors

    # and a simpler, related question: is node A a
    # child of node B?
    def is_child_of(self, candidate_id, allow_self = True):
        # something is often considered a child of itself;
        # we test this first because get_parents will NOT
        # include self in the list, but WILL stop (with
        # an empty list) if self is the candidate
        if self.pk == candidate_id:
            return allow_self
        ancestors = self.get_parents(stop_at_id = candidate_id)
        return len(ancestors) > 0 and ancestors[-1].pk == candidate_id

    # get the root node
    #
    # Similar parent-then-child fetching caveat as
    # get_parents.
    #
    def get_root(self):
        if self.parent_id is None:
            return self
        else:
            return self.get_parents(oldest_first = True)[0]

    # get all siblings
    #
    # NOTE: this is cached on the parent object.
    #
    def get_siblings(self, q = None, order_by = None, select_related = None):
        return self.parent.get_children(q = q, order_by = order_by, select_related = select_related)
        
    # get immediate siblings
    #
    # Do these only if you need exactly one of these; if
    # you need more than one, it's likely more efficient
    # to use get_siblings() and work through the list.
    #
    def get_previous(self):
        return self.__class__.objects.filter(
                display_order_lt = self.display_order,
            ).order_by('-display_order').first()
        
    def get_next(self):
        return self.__class__.objects.filter(
                display_order_gt = self.display_order,
            ).order_by('display_order').first()
        
    # get children
    # 
    # You can specify the number of generations (-1 means
    # "all", but this is dangerous if you're not sure how
    # deep the rabbit hole goes), a Q object to filter
    # children, an alternative ordering, and whether each
    # node should automatically fetch related records.
    #
    # This is a class method, rather than an instance
    # method, because it's very, very common to want to
    # fetch the children for a list of nodes, not just
    # one, and it's much more efficient to fetch them
    # all at once. This method will fetch the children
    # for each node in the nodes list you pass and store
    # them in .children_list on each node; it will not
    # return a useful value.
    #
    @classmethod
    def fetch_children(cls, nodes, generations = 1, q = None, order_by = None, select_related = None):
        from sculpt.model_tools.tools import ModelTools
        
        # default sort order for children
        if order_by is None:
            order_by = [ 'display_order' ]

        # it's often useful to return a dictionary
        # of all the records fetched, indexed by
        # ID, and we can construct it easily as we
        # fetch; we'll also permit nodes to be a
        # dict instead of a list, in which case we
        # will write back to that dict
        if isinstance(nodes, dict):
            all_nodes = nodes
            nodes = nodes.values()

        else:
            all_nodes = dict([ (r.pk,r) for r in nodes ])
        
        # we test for equivalence to zero so that
        # -1 can be passed for "all" (dangerous;
        # if you know you need ALL nodes, not just
        # all the children starting at a particular
        # set of nodes, it's more efficient to use
        # fetch_all_children below)
        while generations != 0:
            generations -= 1
            
            # do the complete fetch
            ModelTools.fetch_related(nodes, 'children', q, order_by, select_related = select_related)
            
            # collect together all the fetched
            # nodes, which are dispersed among
            # the nodes we just queried about
            new_nodes = []
            for n in nodes:
                for ni in range(len(n.children_list)):
                    nn = n.children_list[ni]
                    # it's possible we fetched a
                    # node that we already had;
                    # make sure to de-duplicate
                    if nn.pk in all_nodes:
                        n.children_list[ni] = all_nodes[nn.pk]
                    else:
                        all_nodes[nn.pk] = nn

                    # if we don't already have
                    # a children_list on this node,
                    # we need to look it up in the
                    # next query
                    if not hasattr(nn, 'children_list'):
                        new_nodes.append(nn)
            
            if len(new_nodes) == 0:
                # no children were fetched; we can stop
                break
                
            nodes = new_nodes
            
        return all_nodes

    # and, as a special-case method, we allow fetching
    # the children of one specific node (ourselves)
    def get_children(self, generations = 1, q = None, order_by = None, select_related = None):
        return self.fetch_children([ self ], generations, q, order_by, select_related)

    # sometimes we know we want all the children, and we
    # want to fetch them all at once and then sort them
    # out
    #
    # NOTE: nodes can either be a Model class object, or
    # a list of Model instances; if a Model class object
    # is given, a list of root nodes will be returned
    #
    @classmethod
    def fetch_all_children(cls, nodes, q = None, order_by = None, select_related = None):

        # default sort order for children
        if order_by is None:
            order_by = [ 'display_order' ]
        
        # we need to identify the model to query
        if isinstance(nodes, models.Model):
            node_class = nodes
            node_ids = None
            nodes = None
            
        else:
            # if there are no nodes to fetch children for,
            # we can quit early
            if len(nodes) == 0:
                return
                
            node_class = nodes[0].__class__
            node_ids = [ n.id for n in nodes ]
            
        # fetch all the children
        children = node_class.objects.all()

        if node_ids:
            children = children.exclude(id__in = node_ids)  # don't re-fetch these parents

        if q is not None:
            children = children.filter(q)

        if order_by is not None:
            # a common mistake is to pass a single field name
            # instead of a list; catch this and rework it
            if isinstance(order_by, basestring):
                order_by = [ order_by ]
            children = children.order_by(*order_by)

        if select_related is not None:
            # a common mistake is to pass a single field name
            # instead of a list; catch this and rework it
            if isinstance(select_related, basestring):
                order_by = [ select_related ]
            children = children.select_related(*select_related)
        
        # create a quick index to all the children and
        # (if present) the original parents
        node_index = dict([ (n.id,n) for n in children ])
        if nodes is not None:
            node_index.update(dict([ (n.id,n) for n in nodes ]))    # add the parents to the index

        # now process all the children in order, assigning
        # them to their parents' children_list
        roots = []
        for n in children:
            n.children_list = []
            if n.parent_id is None:
                roots.append(n)
        
        for n in children:
            if n.parent_id is not None and n.parent_id in node_index:
                # only if we've fetched the parent; if we don't have
                # the parent it means the q parameter was too strict
                # and we omitted the parent when fetching "all" the
                # children, which will result in additional queries
                # when fetching the parent later (don't do that)
                
                # set object reference so Django doesn't load
                # duplicates for each child
                n.parent = node_index[n.parent_id]
                
                # add this node to the list of children, in order
                n.parent.children_list.append(n)
        
        return roots



from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.manager import Manager, QuerySet
from sculpt.common import Enumeration

# ModelTools
#
# Django's ORM is pretty good but it has some deficiencies.
# We collect here some functions which are often useful but
# which we do not want to add to the model classes as mix-ins.
#
class ModelTools(object):

    # fetch_related
    #
    # Django's object relational mapping (ORM) system has some
    # very nice automatic related-model fetching. For example, if
    # you create a QuerySet of items with a ForeignKey to another
    # model, when you reference the related model by its property
    # name on an object in the query set, it is fetched from the
    # database and stored in a cache on the QuerySet. Further,
    # you can instruct Django to pre-populate this cache using
    # the select_related() method on the QuerySet; this will not
    # even incur an extra query, as it will be done as a JOIN in
    # SQL so that data comes back all at the same time as the
    # main query.
    #
    # However, working in the other direction is not nearly as
    # efficient. If model A has a ForeignKey to model B, there
    # are plenty of use cases for querying model B, and in a
    # nested loop processing all of the A records linked to each
    # instance of model B. Django DOES provide an automatic
    # QuerySet generator on the model B object, so you can issue
    # a B.related_field.all() to get the records, but it has some
    # limitations:
    #
    #   1. It issues a separate query for each B record. While
    #      the difference in raw Python execution time is not
    #      huge, the additional load on the database server is
    #      and the latency of the extra queries is noticeable.
    #   2. The results are not cached AT ALL; the result is a
    #      new QuerySet that isn't connected to its parent in
    #      any way.
    #
    # To address this weak spot we use this method. Given a
    # QuerySet, a related field name, and optional Q object
    # and sort order, the related records are fetched in one
    # query, then collected together into lists for each record
    # in the original QuerySet.
    #
    # If you do not specify a results field to store the lists
    # in, it defaults to related_field + '_list'.
    #
    # NOTE: returns the SAME (modified) QuerySet.
    # NOTE: only basic customization of the QuerySet is possible;
    # at some point we may need to allow a callable to be passed
    # in that will modify the QuerySet, to allow defer() or
    # extra() or other fun things.
    #
    # NOTE: Django as of v1.4 does offer prefetch_related which
    # provides similar functionality, including nested lookups,
    # but does so at the expense of being able to specify a
    # Q object to filter the results. This function now has a
    # select_related method.
    #
    @classmethod
    def fetch_related(cls, qs, related_field, q = None, order_by = None, results_field = None, id_list = None, select_related = None, fix_reverse_links = True):
        if results_field == None:
            results_field = related_field + '_list'

        # determine which IDs to include in our related record
        # query; exclude None values, which Django uses for
        # records which are unsaved (and thus can't have any
        # related records in the database)
        if id_list == None:
            id_list = [ r.id for r in qs if r.id != None ]

        # Since we might exit early, we go ahead and create the
        # empty lists for each record in the QuerySet now.
        for r in qs:
            setattr(r, results_field, [])

        if len(id_list) == 0:
            # there are no records with IDs to fetch; DO NOT
            # actually issue such a pointless query
            # (more importantly, we can't tell what model to query)
            return qs

        # determine the model name of the related field

        # This is a bit dicey, because we have to look inside the
        # original ForeignKey object on the base class object;
        # in order to do that, we have to get that class object.
        # We can do this in one of two ways:
        #
        #   1. Peek inside the QuerySet and pull it from there.
        #   2. Look at the first object in the QuerySet and
        #      get its class directly.
        #
        # We use #2, for these reasons:
        #
        #   a. We want to muck around with Django internals as
        #      little as possible. They're internal for a reason,
        #      and we run the risk of broken code when we update
        #      Django.
        #   b. We'd like to be able to work with a list, not
        #      just a QuerySet. QuerySets behave like lists in
        #      many contexts and it's very convenient to have
        #      all helper functions work with either.
        #
        # One quirk is that unsaved objects in a list will have
        # an ID of None (this is how Django decides if it is new
        # or not). We can still get the class name, but we don't
        # include these in the list of IDs we query for.

        # take the first object and get the relationship, then
        # extract the model class object and the reverse field
        # name (which we need to find the parent object)
        relationship = getattr(qs[0].__class__, related_field).related
        related_model = relationship.model
        related_model_field_name = relationship.field.name
        related_model_field_id = relationship.field.attname

        # fetch all the related records

        # we can't use Django's in_bulk() here because we want
        # to return records in correct sorted order, and the
        # in_bulk() returns a dict; we just use a normal
        # QuerySet
        # and the odd **{} is idiomatic Python for constructing
        # function parameter names on the fly
        rqs = related_model.objects.filter(**{ related_model_field_name +'_id__in': id_list })

        if q is not None:
            rqs = rqs.filter(q)

        if order_by is not None:
            # a common mistake is to pass a single field name
            # instead of a list; catch this and rework it
            if isinstance(order_by, basestring):
                order_by = [ order_by ]
            rqs = rqs.order_by(*order_by)

        if select_related is not None:
            # a common mistake is to pass a single field name
            # instead of a list; catch this and rework it
            if isinstance(select_related, basestring):
                order_by = [ select_related ]
            rqs = rqs.select_related(*select_related)

        # place each related record with its proper parent

        # to do this efficiently, we need a map of parent ID
        # to object; this dense bit of idiomatic Python does it
        qs_map = dict([ (r.id,r) for r in qs ])

        # now sift each related record (rr)
        # while we are at it, we will (if asked) link the
        # child record back to its parent, because Django
        # doesn't know when it fetches the related records
        # that we already have the original record
        for rr in rqs:
            r = qs_map[getattr(rr, related_model_field_id)] # get parent record
            getattr(r, results_field).append(rr)            # append results to list
            if fix_reverse_links:
                setattr(rr, related_model_field_name, r)    # link related object back to its parent

        return qs

    # update_or_create
    #
    # Django offers a useful get_or_create method which will
    # either fetch a record or, if it doesn't exist, create it
    # with a given set of defaults. This is good, but there is
    # also a very common pattern of fetch-update-save which can
    # benefit from this create-if-missing behavior.
    #
    # Pass 'defaults' as fields to use when creating missing
    # records. Pass 'updates' as fields to replace when updating
    # (these will automatically be applied on top of defaults).
    # Everything else is assumed to be a field lookup, and if
    # it does not contain __ it will be added to the defaults.
    #
    # Returns (record, created, updated)
    #
    # NOTE: Django 1.7 offers an update_or_create method but it
    # doesn't separate defaults from updates, so there's no way
    # to set default values for new records without forcing an
    # existing record to be reset. This is dumb and short-
    # sighted.
    #
    @classmethod
    def update_or_create(cls, model_or_manager, *args, **kwargs):
        # these parameters can't be listed in the formal list
        # or Python will attempt to fill them with positional
        # parameters, which we DO NOT WANT.
        defaults = kwargs.pop('defaults', {})
        updates = kwargs.pop('updates', {})

        # merge updates onto defaults
        defaults.update(updates)

        # normally we expect a model object as our first parameter,
        # but we might get a manager (e.g. related field manager)
        # to simplify working with related objects
        if isinstance(model_or_manager, Manager):
            manager = model_or_manager
        else:
            manager = model_or_manager.objects  # use default manager

        # fetch or create the record
        record, created = manager.get_or_create(*args, defaults = defaults, **kwargs)

        # if we didn't create it, we need to finish the update
        # NOTE: we don't save the record unless we modify at
        # least one field
        dirty = False
        if not created:
            for k,v in updates.iteritems():
                if getattr(record, k) != v:
                    dirty = True
                    setattr(record, k, v)
            if dirty:
                record.save()

        return (record, created, dirty)

    # set_and_track_dirty
    # 
    # Along with update_or_create, it's also often useful
    # to be able to set fields on a record from a dictionary
    # but keep track of whether any of the fields resulted
    # in a change; later, we can save the record if it's
    # "dirty"
    #
    # We make a few assumptions when we do this. First, we
    # assume that the fields we're setting aren't hidden
    # behind properties, because we're going to compare them
    # directly. Second, we assume that you will reset the
    # dirty field list if you manually save the object.
    #
    # This part of the process, where we just set fields on
    # the record, does NOT touch the database unless you are
    # using deferred fields, in which case we force a fetch
    # of any field we're trying to set.
    #
    # NOTE: we return the record in case you want to save it
    # right away.
    #
    @classmethod
    def set_and_track_dirty(cls, record, attrs):
        # ensure we have a list of dirty fields; note that
        # we don't erase an existing list because it might
        # be dirty fields from a previous invocation
        if not hasattr(record, '_caxiam_dirty_list'):
            record._caxiam_dirty_list = []
        for k,v in attrs.iteritems():
            if not hasattr(record, k):
                raise AttributeError('record type %s does not have attribute %s' % (record.__class__.__name__, k))
            if getattr(record, k) != v:
                setattr(record, k, v)
                record._caxiam_dirty_list.append(k)
        return record

    # save the record, but only the dirty fields
    # NOTE: returns the list of fields updated
    @classmethod
    def save_if_dirty(cls, record):
        if record.id == None:
            # although is_dirty() returns True if the ID is
            # None, we need to special-case our processing
            # to make sure we don't pass an update_fields list
            # for new records
            record.save()
            
            # create an empty dirty list and report we saved
            # everything
            record._caxiam_dirty_list = []
            return [ f.name for f in record._meta.fields ]  # messy Django-internals stuff
            
        if cls.is_dirty(record):
            # we have a dirty list and it's not empty
            record.save(update_fields = record._caxiam_dirty_list)
            
            # preserve the dirty list as we're going to wipe
            # it out
            dirty_list = record._caxiam_dirty_list
            record._caxiam_dirty_list = []
            return dirty_list

        # nothing is dirty, don't save
        return []

    # determine whether a record was marked dirty
    # any record without an ID is automatically dirty (it
    # has never been saved) but if it has an ID, it has to
    # be explicitly marked dirty by set_and_track_dirty
    @classmethod
    def is_dirty(cls, record):
        return record.id == None or (hasattr(record, '_caxiam_dirty_list') and record._caxiam_dirty_list)

    # given a model class and a dictionary of parameters,
    # pass all the ones that are valid field names into the
    # model constructor to create an unsaved record, then
    # set all the unused parameters as additional attributes
    # on the record
    #
    # this is useful when you have a collection of data and
    # you want to create objects with valid parameters, but
    # still have the extra data available on the current
    # request that won't be saved
    #
    # Unfortunately Django's meta-class magic means things
    # assigned to the class don't exist as fields on the
    # class unless they're foreign keys; they're collected
    # up and added to _meta.fields instead. Fetching them
    # from there is a bit dirty.
    #
    @classmethod
    def create_with_extra(cls, model_class, attrs):

        # look up the fieldnames from _meta
        model_fields = frozenset([ unicode(f.name) for f in model_class._meta.fields ])

        # split the attrs into two dicts
        model_kwargs = {}
        extra_attrs = {}
        for k,v in attrs.iteritems():
            k = unicode(k)
            if k in model_fields:
                model_kwargs[k] = v
            else:
                extra_attrs[k] = v
                
        # create the model object
        m = model_class(**model_kwargs)

        # apply the extra attributes
        for k,v in extra_attrs.iteritems():
            setattr(m, k, v)

        # return the result
        return m

# OneToOneReverse
#
# Django offers a OneToOneField which is convenient (it's a
# ForeignKey with a unique constraint) but the two sides of
# the relationship don't behave the same. If you access the
# side where OneToOne is defined, and no related record
# exists, a None value is given. If you access the other
# side and no record exists, an ObjectDoesNotExist exception
# is thrown. Oops.
#
# This method allows you to define a property that works on
# the reverse side as it does on the forward side. On the
# forward side, point the related_name to a private field;
# then on the reverse side, define the public name with this
# method pointed at the private field.
#
# Under the hood, this returns a customized function.
#
# NOTE: this now CACHES the reverse lookup. You can clear
# the cache by deleting the attribute, which has _cache
# appended to the fieldname.
#
# EXAMPLE USAGE:
# class A():
#   models.OneToOneField(B, related_name="_a")
#
# class B():
#   a = property(OneToOneReverse('_a'))
#
def OneToOneReverse(fieldname):
    def inner(self):
        cachename = fieldname + '_cache'
        if not hasattr(self, cachename):
            # no cached lookup
            try:
                f = getattr(self, fieldname)
            except ObjectDoesNotExist:
                f = None
            setattr(self, cachename, f)
        return getattr(self, cachename)
    return inner

# set database transaction isolation modes
#
# NOTE: this is the low-level function, not the view
# decorator. See caxiam.decorators.set_isolation_mode
# for that.
#
# Django offers no support for setting the database
# transaction isolation mode, defaulting to whatever the
# database is configured for. The default for MySQL is
# repeatable read, which effectively freezes the visible
# data to its state at the start of the transaction; this
# is reasonable for most situations. However when dealing
# with concurrent updates of data where thresholds are
# important, race conditions apply. Use this function to
# set the isolation mode for the next transaction.
#
# Consider this example. Two players, A and B, are playing
# a turn-based game with simultaneous moves. Each submits
# their move at roughly the same time.
#
#       A                       B
# 1     begin txn
# 2                             begin txn
# 3     attempt game lock
# 4     receive game lock
# 5                             attempt game lock
# 6     write move
# 7     count moves
# 8     1 move: turn not over
# 9     end txn / release lock
# 10                            receive game lock
# 11                            write move
# 12                            count moves
# 13                            1 move: turn not over
# 14                            end txn / release lock
#
# The answer at step 13 is 1, not 2, because B's transaction
# started at 2, freezing B's view of the database to that
# point, due to repeatable read. B can't see the write at
# step 8 until both A's transaction completes (committing it
# to the database for other processes to read) and B's
# transaction completes (allowing it to start a new
# transaction with an updated snapshot).
#
# In this circumstance, the process needs to use the
# READ_COMMITTED to force each read to fetch a more current
# snapshot, or break the process into multiple transactions.
#
ISOLATION_MODES = Enumeration(
        (0, 'REPEATABLE_READ'),
        (1, 'READ_COMMITTED'),
        (2, 'READ_UNCOMMITTED'),
        (3, 'SERIALIZABLE'),
    )

def set_isolation_mode(isolation_mode):
    # So Django is messed up. Python says new queries should
    # open a transaction and LEAVE IT OPEN for the app to
    # commit or roll back. Django interprets this as "commit
    # on write or at request end" because manually committing
    # every transaction is not friendly to devs. Adding true
    # autocommit support would be better because the way it is
    # now, any innocent read will leave an open transaction.
    #
    # This has major problems for setting isolation modes,
    # which can only be done OUTSIDE of a transaction unless
    # you're going to set it for the whole session. We don't
    # want to set it for the whole session as we want it to
    # revert after the next transaction to its app-configured
    # default. And, if our app is using sessions, Django has
    # automatically performed a read, leaving an open
    # transaction.
    #
    # To work around this, we go ahead and COMMIT any open
    # transaction. This is grossly inelegant and if our
    # assumptions about why there is an open transaction are
    # wrong, this code will fail. Worse, when Django gets
    # around to fixing their quirky transaction behavior
    # (which looks like it will be in 1.6) this fix may then
    # break because there is no open transaction.
    
    # NOTE: we do not use Django's parameter quoting because
    # we are inserting SQL keywords, not values, into the
    # statement and we don't want Django/MySQLdb quoting them
    from django.db import connection
    cursor = connection.cursor()
    if settings.CAXIAM_DUMP_SQL:
        print "[pid:%d]" % os.getpid(), 'SETTING ISOLATION MODE', ISOLATION_MODES.get_label(isolation_mode)
    cursor.execute(
            'commit; set transaction isolation level ' + ISOLATION_MODES.get_label(isolation_mode).replace('_', ' '),
            []
        )
    cursor.fetchone()




from django.conf import settings
from django.http import HttpResponseRedirect

# LoginRequiredMixin
# This mixin allows our views to be auth controlled and we
# can change the settings from the urls.py level.
#
# A url might look like: 
# url(r'^backoffice/members/$', <View>.as_view(template_name = ..., login_required = True, login_redirect_location = '/backoffice/login/'))
#
# You can also set global defaults for your entire
# application in settings: 
#
# LOGIN_REQUIRED_DEFAULT
# LOGIN_SESSION_KEY
# LOGIN_REDIRECT_LOCATION
#
class LoginRequiredMixin(object):

    # Whether the login is required for this view.
    # True for yes, None for maybe, false for None required
    login_required = settings.LOGIN_REQUIRED_DEFAULT 
    
    # Session Key used when logged in to determine if they
    # are logged in or not; you should not need to override
    # this as it's likely set site-wide, but if you're using
    # two completely separate user systems (e.g. customers,
    # back office) then you should subclass the mixin, set
    # it in the subclass, and then apply the mixin subclass to
    # your views, rather than override this in each of your
    # views.
    login_session_key = settings.LOGIN_SESSION_KEY_DEFAULT
    
    # Redirect Location to go when ensuring fails (see above
    # comment for login_session_key, it applies here)
    login_redirect_location = settings.LOGIN_REDIRECT_LOCATION_DEFAULT

    # What to do when you aren't logged in for a GET response
    # By default it will do an HTTP response redirect to the
    # redirect location
    def _get_if_not_logged_in(self, *args, **kwargs):
        return HttpResponseRedirect(self.login_redirect_location)

    # What to do when you aren't logged in for a POST response
    # By default it will do whatever _get_if_not_logged_in() does,
    # but a reasonable alternative might be to return an AJAX
    # redirect (see AjaxLoginRequiredMixin)
    def _post_if_not_logged_in(self, *args, **kwargs):
        return self._get_if_not_logged_in(*args, **kwargs)

    # This will check if you are logged in currently and
    # dispatch a type response if you are not.
    def _check_login(self, request, *args, **kwargs):
        if request.session.get(self.login_session_key) is None:
            method_name = '_'+request.method.lower()+'_if_not_logged_in'
            return getattr(self, method_name)(request = request, *args, **kwargs)

    # WARNING: Be very careful when overrideing this function.
    # This almost always needs to run first, so call super()
    # before doing something different.
    #
    # If the setting LOGIN_REQUIRED_DEFAULT is set to True,
    # or if you have said login_required is True in the
    # urls.py then it will check to see if you have a valid
    # login. If it does then continue normally, if not it will
    # try to return an appropriate redirect-ish response.
    #
    def dispatch(self, request, *args, **kwargs):
        if self.login_required == True:
            response = self._check_login(request = request, *args,**kwargs)
            if response:
                return response
        return super(LoginRequiredMixin, self).dispatch(request = request, *args, **kwargs)


# AjaxLoginRequiredMixin
#
# Almost identical to LoginRequiredMixin, but assumes a POST
# request is AJAX and thus needs an AJAX redirect if the user
# is not logged in.
#
class AjaxLoginRequiredMixin(LoginRequiredMixin):

    def _post_if_not_logged_in(self, *args, **kwargs):
        from sculpt.ajax.responses import AjaxRedirectResponse
        return AjaxRedirectResponse(self.login_redirect_location)



import base64
import hashlib
import hmac
import numbers

# Create a Hash given a list of items
class ModelHashGenerator(object):

    @classmethod
    def generate_hash(klass, cls, hashkey, *args):       
        counter = 0
        while True:
            counter += 1
            encoded_hash = klass.generate_hash_core(counter, hashkey, *args)
            if encoded_hash[0] in '-_':
                # we don't want user hashes that start with these
                continue
            if cls.objects.filter(hash = encoded_hash).count() > 0:
                # this hash is already in use
                continue
                
            # otherwise the hash is acceptable
            return encoded_hash

    @classmethod
    def generate_hash_core(cls, counter, hashkey, *args):
        raw_hash = hmac.new(hashkey, repr(args) + str(counter), hashlib.sha256).digest()
        encoded_hash = base64.urlsafe_b64encode(raw_hash)[:43]  # strips always-present trailing =
        return encoded_hash
        


from django.db import models
from django.contrib.auth.models import AbstractBaseUser
from django.utils import timezone

from sculpt.ajax.enumerations import ISO_COUNTRIES
from sculpt.common import Enumeration, EnumerationData
from sculpt.model_tools.base import AbstractAutoHash
from sculpt.model_tools.mixins import AutoHashMixin, LoginMixin, OverridableChoicesMixin

import datetime

# Django's auth module contains a User model. With
# recent versions it's finally possible to work around
# its limitations, but if you don't want to use and
# extend Django's built-in User system, you need to
# build user models based on something else. These
# abstract base classes provide one such starting
# point.
#
# Reasons you might not want to use Django's User
# models:
#
# 1. Django's user system is linked to the Django
# admin app. This means your app's users are in the
# same system as the tool that offers extremely low-
# level access to your database. Obviously you protect
# access to this, but ask yourself which is a more
# likely and harder-to-spot programming mistake:
# failing to notice that a session is linked to a
# user AT ALL, or failing to correctly parse which
# level of access a particular user happens to have?
#
# 2. You might have a lot of users and you just don't
# want to try to put those users into Django's User
# database, where it's harder to figure out who is
# supposed to have access and who is not.
#
# 3. You are building an app for a client, and when
# you are done, you need to hand off user management
# for the application to the client, but you largely
# do not expect them to be mucking around with the
# Django admin tool. It's much easier to build tools
# that work with completely separate models than to
# make tools that check whether they're allowed to
# modify specific records.
#
# These are brief summaries and this isn't intended
# as a universal condemnation of full integration
# with Django's User system. The point is to get you
# to think about your user management rather than
# adopt a default solution that might work against
# you later.
#
# If you choose to create a separate user system,
# two common patterns are supported. "Simple" app
# users assume that each user has exactly one username
# and password that can be used to authenticate,
# and this is derived from Django's AbstractBaseUser
# in order to pick up its password-handling ability.
# The other pattern separates credentials from the
# user, so that you can support multiple authentication
# schemes for the same account (e.g. initially creating
# an account with a username/password and then later
# linking it to a Facebook account). It's more work
# but more flexible.
#
# No matter which you choose, they will pick up both
# the LoginMixin and AbstractAutoHash mixins.

""" SIMPLE APP USER SETUP """

# AbstractSimpleAppUser
#
# REQUIRED OVERRIDES:
#   - AUTOHASH_SECRET - a unique string (AbstractAutoHash)
#
# OPTIONAL OVERRIDES:
#   - AUTOHASH_FIELDS - fields used to generate hash (AbstractAutoHash)
#   - CHECK_PASSWORD_METHOD - the name of the function used to check
#       passwords (AbstractBaseUser)
#   - REQUIRED_FIELDS - indicates which fields are required
#       (AbstractBaseUser)
#
# Description:
#   This is to abstract out the most common features
#   of an App User. This class represents the most
#   basic logins, where the only way to authenticate
#   would be a username and password (always). It is
#   easier to implement than the multiple-auth scheme
#   user.
#
#   This model REQUIRES that usernames be unique. If
#   you have a situation that allows users to be
#   deleted, you need to have a strategy for dealing
#   with usernames if you do not want the default
#   behavior this restriction implies (that once used,
#   a username is gone forever).
#
class AbstractSimpleAppUser(LoginMixin, AbstractAutoHash, AbstractBaseUser):
    class Meta(AbstractBaseUser.Meta):
        abstract = True

    # AbstractBaseUser settings
    CHECK_PASSWORD_METHOD = 'check_password'
    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = [ 'username' ]

    # AutoHashMixin settings
    AUTOHASH_FIELDS = [ 'username' ]

    # actual fields on this model
    username = models.CharField(max_length = 32, unique = True)

    # validate a username and password, returning
    # the matched user or None
    # NOTE: it always filters on the value from USERNAME_FIELD
    @classmethod
    def authenticate(cls, username, password):
        user = cls.objects.filter(**{ cls.USERNAME_FIELD: username }).first()
        if user is not None and hasattr(user, cls.CHECK_PASSWORD_METHOD) and getattr(user,cls.CHECK_PASSWORD_METHOD)(password):
            return user


""" COMPLEX APP USER SETUP """

# AbstractAppUser
#
# REQUIRED OVERRIDES:
#   - AUTOHASH_SECRET - a unique string (AbstractAutoHash)
#   - authenticate - method used to authenticate a user,
#       typically by knowing which authentication method
#       is expected, and checking against credentials of
#       that type
#
# OPTIONAL OVERRIDES:
#   - AUTOHASH_FIELDS - fields used to generate hash (AbstractAutoHash)
#
# Description:
#   Unlike AbstractSimpleAppUser, this class is for users
#   that could be authenticated by any of several means.
#   For example, by username and password, or by Facebook
#   token, or by Twitter token, or by unique device ID.
#   Most of the heavy-lifting is done in the Credentials
#   class, leaving this an empty shell. You will still want
#   to add fields that describe your users.
#
class AbstractAppUser(LoginMixin, AbstractAutoHash, models.Model):
    class Meta(object):
        abstract = True

    @classmethod
    def authenticate(cls, type, *args, **kwargs):
        """ MUST OVERRIDE """
        raise Exception("this function must be overridden")


# AbstractAppUserCredential
#
# SUGGESTED CODE IN SUBCLASS:
#   - appuser = models.ForeignKey(<class Child(AbstractAppUser)>, related_name = 'credentials')
#
# REQUIRED OVERRIDES:
#   - CREDENTIAL_TYPES - Enumeration or Tuple
#   - authenticate() - function - each instance of a credential should be able to authenticate itself to see if it's valid.
#
# Description:
#   This represents a single set of credentials for a user,
#   and connects to the code for validating those
#   credentials.
#
class AbstractAppUserCredential(OverridableChoicesMixin, models.Model):
    class Meta(object):
        abstract = True

    CREDENTIAL_TYPES = Enumeration() # Expects to be enumeration

    data1 = models.CharField(max_length = 255, blank = True, null = True)  # typically a username or user ID
    data2 = models.CharField(max_length = 255, blank = True, null = True)  # typically a hashed password or auth token
    credential_type = models.IntegerField(choices = CREDENTIAL_TYPES.choices)
    credential_type_data = property(EnumerationData('CREDENTIAL_TYPES', 'credential_type'))

    # We are running the init function to override the credential type choices
    def __init__(self, *args, **kwargs):
        super(AbstractAppUserCredential, self).__init__(*args, **kwargs)
        self._set_field_choices(field_name = 'credential_type', choices = self.CREDENTIAL_TYPES.choices)

    # This function is used to identify if your current user is permitted to do an action.
    # Most commonly used in login to ensure that a user is allowed to come into the site
    # RETURN Boolean
    def authenticate(self, user):
        """ MUST OVERRIDE """
        raise Exception("this function must be overridden")


# Contact Information
#
# This is a very common pattern, so we set these up once and re-use
# them in multiple apps. Addresses should always be prepared to handle
# international locations.
#
# Generally when you create a concrete implementation of this class
# you will add foreign keys to the things that may have contact info
# associated with them (e.g. something derived from AbstractAppUser).
# However, if you are absolutely certain that a record only needs one
# contact info object, you can construct the relationship from the
# other direction.
#
class AbstractContactInfo(OverridableChoicesMixin, models.Model):
    class Meta(object):
        abstract = True
    
    # does this address have a name or label?
    # (if not, we use the address type)
    title = models.CharField(max_length = 30, blank = True, null = True)

    # by default we will take the country list as the ISO country list
    COUNTRIES = ISO_COUNTRIES
    
    # all fields optional
    address1 = models.CharField(max_length = 100, blank = True, null = True)
    address2 = models.CharField(max_length = 100, blank = True, null = True)
    address3 = models.CharField(max_length = 100, blank = True, null = True)    # primarily international
    city     = models.CharField(max_length = 100, blank = True, null = True)
    state    = models.CharField(max_length =  50, blank = True, null = True)    # or province
    zip      = models.CharField(max_length =  50, blank = True, null = True)    # or postal code
    country  = models.CharField(max_length =   2, blank = True, null = True, choices = COUNTRIES.choices, default = 'US')    # ISO 3166-1-alpha-2; see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    country_data = property(EnumerationData('COUNTRIES','country'))

    # what type of address is it?
    # NOTE: actual concrete implementations may want to override
    # this list with a subset/superset
    ADDRESS_TYPES = Enumeration(
            (0, 'BILLING', 'Billing'),
            (1, 'SHIPPING', 'Shipping'),
        )
    address_type = models.IntegerField(choices = ADDRESS_TYPES.choices, default = 0)
    address_type_data = property(EnumerationData('ADDRESS_TYPES','address_type'))

    # what is the display order/preference?
    # we use this to determine a "best" address for a person
    display_order = models.IntegerField(default = 0)

    # handle overridable enumerations
    def __init__(self, *args, **kwargs):
        super(AbstractContactInfo, self).__init__(*args, **kwargs)
        self._set_field_choices(field_name = 'country', choices = self.COUNTRIES)
        self._set_field_choices(field_name = 'address_type', choices = self.ADDRESS_TYPES)


# Phone Number
#
# You will need to add foreign keys when you create a concrete
# implementation. Depending on your data model, you may want to
# create foreign keys to something derived from AbstractContactInfo
# (if phone numbers are associated most closely with an address)
# or something from AbstractAppUser (if phone numbers are associated
# most closely with a person).
#
class AbstractPhoneNumber(OverridableChoicesMixin, models.Model):
    class Meta(object):
        abstract = True

    # does this number have a name or label?
    # (if not, we use the number type)
    title = models.CharField(max_length = 30, blank = True, null = True)
    
    # what is the actual number?
    number = models.CharField(max_length = 30)
    
    # what type of number is it?
    # NOTE: actual concrete implementations may want to override
    # this list with a subset
    NUMBER_TYPES = Enumeration(
            (0, 'UNKNOWN', 'Unknown'),
            (1, 'HOME', 'Home'),
            (2, 'WORK', 'Work'),
            (3, 'CELL', 'Cell/Mobile'),
            (4, 'FAX', 'Fax'),
        )
    number_type = models.IntegerField(choices = NUMBER_TYPES.choices, default = NUMBER_TYPES.UNKNOWN)
    number_type_data = property(EnumerationData('NUMBER_TYPES','number_type'))
    
    # are there additional notes about when this number can
    # be called or special instructions?
    notes = models.CharField(max_length = 100, blank = True, null = True)

    # what is the display order/preference?
    # we use this to determine a "best" number for a person
    display_order = models.IntegerField(default = 0)

    # handle overridable enumerations
    def __init__(self, *args, **kwargs):
        super(AbstractPhoneNumber, self).__init__(*args, **kwargs)
        self._set_field_choices(field_name = 'number_type', choices = self.NUMBER_TYPES)


# a base email address class that can track its validation
# state; derive from the class and provide any additional
# foreign keys required for your app
class AbstractEmail(OverridableChoicesMixin, models.Model):
    class Meta(object):
        abstract = True

    VERIFICATION_STATES = Enumeration(
            (-3, 'BLOCKED'),       ## Owner opted out of all emails (**** needs implementing)
            (-2, 'VALID'),         ## Known Good
            (-1, 'INVALID'),       ## Known Bad
            ( 0, 'UNVERIFIED'),    ## link sent, awaiting response
            # ... more in-progress unverified states
        )

    address = models.CharField(max_length = 254, db_index = True)    # intentionally long, but MUST BE INDEXED
    status = models.IntegerField(choices = VERIFICATION_STATES.choices)
    status_data = property(EnumerationData('VERIFICATION_STATES','status'))

    # when is this record vreated?
    date_created = models.DateTimeField(auto_now = False, auto_now_add = False, default = datetime.datetime.utcnow)
    # when is this validated by the end user?
    date_validated  = models.DateTimeField(auto_now = False, auto_now_add = False, blank = True, null = True)

    # handle overridable enumerations
    def __init__(self, *args, **kwargs):
        super(AbstractEmail, self).__init__(*args, **kwargs)
        self._set_field_choices(field_name = 'status', choices = self.VERIFICATION_STATES)

    # a helper method to automatically mark an email address as
    # verified and mark all other copies of the same address as
    # invalid
    #
    # NOTE: if your derived class overrides VERIFICATION_STATES
    # such that VALID is renamed or renumbered, you will want
    # to override this method; better yet, don't rename VALID
    #
    # NOTE: unless you pass save = False, this will update the
    # database
    #
    def mark_as_valid(self, save = True):
        # first, update ourselves
        self.status = self.VERIFICATION_STATES.VALID
        self.date_validated = datetime.datetime.utcnow()
        if save:
            self.save(update_fields = [ 'status', 'date_validated' ])

        # second, look for other records in UNVERIFIED
        self.__class__.objects.filter(
                address = self.address, status = self.VERIFICATION_STATES.UNVERIFIED,
            ).exclude(
                id = self.id,
            ).update(
                status = self.VERIFICATION_STATES.INVALID,
            )


# AbstractMultipleEmail
#
# Although the base AbstractEmail class is designed so that it
# can be used as a one-to-many (multiple email addresses per
# other thing, such as user) there's still an assumption that
# there is probably "one" email address, and we support multiple
# in order to ease the transition. In cases where we have true
# multiple-email-address support we need to have labels and
# preferences for those email addresses, similar to how we have
# them for addresses and phone numbers.
#
class AbstractMultipleEmail(AbstractEmail):
    class Meta(object):
        abstract = True

    # does this number have a name or label?
    title = models.CharField(max_length = 30, blank = True, null = True)
    
    # what is the display order/preference?
    # we use this to determine a "best" address for a person
    display_order = models.IntegerField(default = 0)


# Web Sites
#
# Sometimes we want to associate web site address(es) with
# users or entities. This is a base class for doing so.
#
class AbstractWebSite(models.Model):
    class Meta(object):
        abstract = True

    # does this number have a name or label?
    title = models.CharField(max_length = 30, blank = True, null = True)

    # what is the actual web site?
    # NOTE: we don't make this a Django URLField because
    # we don't want their validation rules.
    url = models.CharField(max_length = 250)
    
    # what is the display order/preference?
    # we use this to determine a "best" address for a person
    display_order = models.IntegerField(default = 0)




from django.db import models
from django.db.models import Q
from django.contrib.auth.models import AbstractBaseUser
from django.utils import timezone

from sculpt.common import Enumeration
from sculpt.model_tools.mixins import AutoHashMixin

# AbstractAutoHash
#
# A thin wrapper that creates a model that includes
# AutoHashMixin. AutoHashMixin doesn't define the
# hash field, but expects the application to do so;
# this is the minimum required for that to function.
#
class AbstractAutoHash(AutoHashMixin, models.Model):
    class Meta(object):
        abstract = True

    # AutoHashMixin requirements
    AUTOHASH_FIELDS = []
    AUTOHASH_SECRET = None

    hash = models.CharField(max_length = 43, unique = True, blank = True, null = True) 


# AbstractSoftDelete
#
# If you inherit from this instead of (or in addition to)
# models.Model, regular .delete() actions will be intercepted
# and will instead update a date_deleted field.
#
# THIS WILL NOT PREVENT RECORDS FROM BEING DELETED. They
# can still be deleted via QuerySet.delete(), by foreign key
# cascade delete, or by a database trigger (horrors). To
# fully prevent records in a specific table from being deleted,
# delete permission must be revoked from the database user
# that the web app uses. That would be the nuclear option.
#
# The primary purpose of this is to simplify soft-delete
# management.
#
class AbstractSoftDelete(models.Model):
    class Meta(object):
        abstract = True

    date_deleted = models.DateTimeField(blank = True, null = True, db_index = True)

    def delete(self):
        raise Exception('This model requires a soft delete.')

    def soft_delete(self, date_deleted = None):
        if date_deleted is None:
            date_deleted = timezone.now()
        self.date_deleted = date_deleted
        self.save(update_fields = ['date_deleted'])

    @property
    def is_deleted(self):
        return self.date_deleted != None

    @classmethod
    def is_not_deleted_q(cls):
        return Q(date_deleted__isnull = True)

    @classmethod
    def is_deleted_q(cls):
        return Q(date_deleted__isnull = False)
    




"""A modest set of tools to work with Django models."""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
# with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
#     long_description = f.read()

setup(
	name='sculpt.model_tools',
	version='0.1',
	description='A modest set of tools to work with Django models.',
	long_description='',
	url='https://github.com/damienjones/sculpt-model-tools',
	author='Damien M. Jones',
	author_email='damien@codesculpture.com',
	license='LGPLv2',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
	], 
	keywords='',
	packages=find_packages(),
	install_requires=[
		'sculpt-common>=0.2',
	],
	# package_data={},
	# data_files=[],
	# entry_points={},
	# console_scripts={},
)
